import json
import os
import struct
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union

from naomi.eeprom import NaomiEEPRom, NaomiEEPRomException


class SettingSizeEnum(Enum):
    UNKNOWN = auto()
    NIBBLE = auto()
    BYTE = auto()


class SettingType(Enum):
    UNKNOWN = auto()
    SYSTEM = auto()
    GAME = auto()


NO_FILE: str = "NO FILE"


class SettingsParseException(Exception):
    def __init__(self, msg: str, filename: str) -> None:
        super().__init__(msg)
        self.filename = filename


class SettingsSaveException(Exception):
    def __init__(self, msg: str, filename: str) -> None:
        super().__init__(msg)
        self.filename = filename


class JSONParseException(Exception):
    def __init__(self, msg: str, context: List[str]) -> None:
        super().__init__(msg)
        self.context = context


class ReadOnlyCondition:
    # A wrapper class to encapsulate that a setting is read-only based on the
    # value of another setting.

    def __init__(self, filename: str, setting: str, name: str, values: List[int], negate: bool) -> None:
        self.filename = filename
        self.setting = setting
        self.name = name
        self.values = values
        self.negate = negate

    def evaluate(self, settings: List["Setting"]) -> bool:
        for setting in settings:
            if setting.name.lower() == self.name.lower():
                if setting.current in self.values:
                    return self.negate
                else:
                    return not self.negate

        raise SettingsSaveException(
            f"The setting \"{self.setting}\" depends on the value for \"{self.name}\" but that setting does not seem to exist! Perhaps you misspelled \"{self.name}\"?",
            self.filename,
        )


class DefaultCondition:
    # A wrapper class to encapsulate one rule for a setting default.

    def __init__(self, name: str, values: List[int], negate: bool, default: int) -> None:
        self.name = name
        self.values = values
        self.negate = negate
        self.default = default


class DefaultConditionGroup:
    # A wrapper class to encapsulate a set of rules for defaulting a setting.

    def __init__(self, filename: str, setting: str, conditions: List[DefaultCondition]) -> None:
        self.filename = filename
        self.setting = setting
        self.conditions = conditions

    def evaluate(self, filename: str, name: str, settings: List["Setting"]) -> int:
        for cond in self.conditions:
            for setting in settings:
                if setting.name.lower() == cond.name.lower():
                    current = setting.current if setting.current is not None else setting.default

                    if cond.negate and current not in cond.values:
                        return cond.default
                    if not cond.negate and current in cond.values:
                        return cond.default

        namelist = list({f'"{c.name}"' for c in self.conditions})
        if len(namelist) > 2:
            namelist = [", ".join(namelist[:-1]), namelist[-1]]
        names = " or ".join(namelist)

        raise SettingsSaveException(
            f"The default for setting \"{self.setting}\" could not be determined! Perhaps you misspelled one of {names}, or you forgot a value?",
            self.filename,
        )


class Setting:
    # A single setting, complete with its name, size (and optional length if
    # the size is a byte), whether it is read-only, the allowed values for
    # the setting and finally the current value (if it has been parsed out
    # of a valid EEPROM file).

    def __init__(
        self,
        name: str,
        size: SettingSizeEnum,
        length: int,
        read_only: Union[bool, ReadOnlyCondition],
        values: Optional[Dict[int, str]] = None,
        current: Optional[int] = None,
        default: Optional[Union[int, DefaultConditionGroup]] = None,
    ) -> None:
        self.name = name
        self.size = size
        self.length = length
        self.read_only = read_only
        self.values = values or {}
        self.current = current
        self.default = default

        if size == SettingSizeEnum.UNKNOWN:
            raise Exception("Logic error!")
        if length > 1 and size != SettingSizeEnum.BYTE:
            raise Exception("Logic error!")

    @staticmethod
    def from_json(filename: str, jsondict: Dict[str, Any], context: List[str]) -> "Setting":
        # Convert JSON dictionary back to a Setting class.
        name = jsondict.get('name')
        if not isinstance(name, str):
            raise JSONParseException(f"\"name\" key in JSON has invalid data \"{name}\"!", context)

        sizestr = jsondict.get('size')
        if not isinstance(sizestr, str):
            raise JSONParseException(f"\"size\" key in JSON has invalid data \"{sizestr}\"!", context)
        try:
            size = SettingSizeEnum[sizestr]
        except KeyError:
            size = SettingSizeEnum.UNKNOWN
        if size == SettingSizeEnum.UNKNOWN:
            raise JSONParseException(f"\"size\" key in JSON has invalid data \"{sizestr}\"!", context)

        length = jsondict.get('length')
        if not isinstance(length, int):
            raise JSONParseException(f"\"length\" key in JSON has invalid data \"{length}\"!", context)

        current = jsondict.get('current')
        if current is not None and not isinstance(current, int):
            raise JSONParseException(f"\"current\" key in JSON has invalid data \"{current}\"!", context)

        valuedict = jsondict.get('values')
        values: Optional[Dict[int, str]] = None
        if valuedict is not None:
            if not isinstance(valuedict, dict):
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{valuedict}\"!", context)
            try:
                values = {int(k): str(v) for (k, v) in valuedict.items()}
            except ValueError:
                pass
            if values is None:
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{valuedict}\"!", context)

        readonlydict = jsondict.get('readonly')
        readonly: Union[bool, ReadOnlyCondition]
        if readonlydict is True:
            readonly = True
        elif readonlydict is False:
            readonly = False
        elif isinstance(readonlydict, dict):
            # Parse out conditional.
            roname = readonlydict.get('name')
            if not isinstance(roname, str):
                raise JSONParseException(f"\"name\" key in JSON has invalid data \"{roname}\"!", [*context, "readonly"])

            rovaluelist = readonlydict.get('values')
            if not isinstance(rovaluelist, list):
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{rovaluelist}\"!", [*context, "readonly"])
            rovalues = None
            try:
                rovalues = [int(k) for k in rovaluelist]
            except ValueError:
                pass
            if rovalues is None:
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{rovaluelist}\"!", [*context, "readonly"])

            ronegate = readonlydict.get('negate')
            if not isinstance(ronegate, bool):
                raise JSONParseException(f"\"negate\" key in JSON has invalid data \"{ronegate}\"!", [*context, "readonly"])

            readonly = ReadOnlyCondition(filename, name, roname, rovalues, ronegate)
        else:
            raise JSONParseException(f"\"readonly\" key in JSON has invalid data \"{readonlydict}\"!", context)

        defaultdict = jsondict.get('default')
        default: Optional[Union[int, DefaultConditionGroup]] = None
        if defaultdict is None or isinstance(defaultdict, int):
            default = defaultdict
        elif isinstance(defaultdict, list):
            # Go through each entry in the list and parse out the conditionals.
            default = DefaultConditionGroup(filename, name, [])

            for i, defaultblob in enumerate(defaultdict):
                # Parse out conditional.
                defname = defaultblob.get('name')
                if not isinstance(defname, str):
                    raise JSONParseException(f"\"name\" key in JSON has invalid data \"{defname}\"!", [*context, f"default[{i}]"])

                defvaluelist = defaultblob.get('values')
                if not isinstance(defvaluelist, list):
                    raise JSONParseException(f"\"values\" key in JSON has invalid data \"{defvaluelist}\"!", [*context, f"default[{i}]"])
                defvalues = None
                try:
                    defvalues = [int(k) for k in defvaluelist]
                except ValueError:
                    pass
                if defvalues is None:
                    raise JSONParseException(f"\"values\" key in JSON has invalid data \"{defvaluelist}\"!", [*context, f"default[{i}]"])

                defnegate = defaultblob.get('negate')
                if not isinstance(defnegate, bool):
                    raise JSONParseException(f"\"negate\" key in JSON has invalid data \"{defnegate}\"!", [*context, f"default[{i}]"])

                defdefault = defaultblob.get('default')
                if not isinstance(defdefault, int):
                    raise JSONParseException(f"\"default\" key in JSON has invalid data \"{defdefault}\"!", [*context, f"default[{i}]"])

                default.conditions.append(DefaultCondition(defname, defvalues, defnegate, defdefault))
        else:
            raise JSONParseException(f"\"default\" key in JSON has invalid data \"{defaultdict}\"!", context)

        return Setting(
            name,
            size,
            length,
            readonly,
            values,
            current,
            default,
        )

    def to_json(self) -> Dict[str, Any]:
        jdict = {
            'name': self.name,
            'size': self.size.name,
            'length': self.length,
            'values': self.values,
            'current': self.current,
        }

        if self.read_only is True:
            jdict['readonly'] = True
        elif self.read_only is False:
            jdict['readonly'] = False
        elif isinstance(self.read_only, ReadOnlyCondition):
            jdict['readonly'] = {
                "name": self.read_only.name,
                "values": self.read_only.values,
                "negate": self.read_only.negate,
            }

        if self.default is None or isinstance(self.default, int):
            jdict['default'] = self.default
        elif isinstance(self.default, DefaultConditionGroup):
            jdict['default'] = [
                {
                    "name": cond.name,
                    "values": cond.values,
                    "default": cond.default,
                    "negate": cond.negate,
                }
                for cond in self.default.conditions
            ]
        return jdict

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


class Settings:
    # A collection of settings as well as the type of settings this is (game versus
    # system). This is also responsible for parsing and creating sections in an actual
    # EEPROM file based on the settings themselves.

    def __init__(self, filename: str, settings: List[Setting], type: SettingType = SettingType.UNKNOWN) -> None:
        self.filename = filename
        self.settings = settings
        self.type = type

    @staticmethod
    def from_config(type: SettingType, config: "SettingsConfig", eeprom: NaomiEEPRom) -> "Settings":
        settings = config.settings
        location = 0
        halves = 0

        if type == SettingType.SYSTEM:
            data = eeprom.system
        elif type == SettingType.GAME:
            data = eeprom.game
        else:
            raise Exception(f"Cannot load settings with a config of type {type.name}!")

        for setting in settings:
            if setting.size == SettingSizeEnum.NIBBLE:
                try:
                    if halves == 0:
                        setting.current = (data[location] >> 4) & 0xF
                    else:
                        setting.current = data[location] & 0xF
                except NaomiEEPRomException:
                    # Out of range, ignore this read.
                    pass

                if halves == 0:
                    halves = 1
                else:
                    halves = 0
                    location += 1
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsParseException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", config.filename)

                try:
                    if setting.length == 1:
                        setting.current = struct.unpack("<B", data[location:(location + 1)])[0]
                    elif setting.length == 2:
                        setting.current = struct.unpack("<H", data[location:(location + 2)])[0]
                    elif setting.length == 4:
                        setting.current = struct.unpack("<I", data[location:(location + 4)])[0]
                    else:
                        raise SettingsParseException(f"Cannot parse setting \"{setting.name}\" with unrecognized size \"{setting.length}\"!", config.filename)
                except NaomiEEPRomException:
                    # Out of range, ignore this read.
                    pass

                location += setting.length

        final_settings = Settings(config.filename, settings, type=type)
        if type == SettingType.SYSTEM:
            expected = 16
            sectype = "system"
        elif type == SettingType.GAME:
            expected = eeprom.length
            sectype = "game"
        else:
            raise Exception(f"Cannot load settings with a config of type {type.name}!")

        if final_settings.length != expected:
            raise SettingsParseException(
                f"Unexpected final size of {sectype} section, expected {expected} bytes but definition file covers {final_settings.length} bytes!",
                config.filename,
            )
        return final_settings

    @property
    def length(self) -> int:
        halves = 0
        length = 0
        for setting in self.settings:
            if setting.size == SettingSizeEnum.NIBBLE:
                # Update our length.
                if halves == 0:
                    halves = 1
                else:
                    halves = 0
                    length += 1

            elif setting.size == SettingSizeEnum.BYTE:
                # First, make sure we aren't in a pending nibble state.
                if halves != 0:
                    raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", self.filename)

                if setting.length not in {1, 2, 4}:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", self.filename)

                # Update our length.
                length += setting.length

        return length

    @staticmethod
    def from_json(type: SettingType, config: "SettingsConfig", jsondict: Dict[str, Any], context: List[str]) -> "Settings":
        # First, parse out the keys that we know we need.
        typestr = jsondict.get('type')
        if not isinstance(typestr, str) or typestr != type.name:
            raise JSONParseException(f"\"type\" key in JSON has invalid data \"{typestr}\"!", context)
        filename = jsondict.get('filename')
        if not isinstance(filename, str) or filename != config.filename:
            raise JSONParseException(f"\"filename\" key in JSON has invalid data \"{filename}\"!", context)
        settings = jsondict.get('settings')
        if not isinstance(settings, list):
            raise JSONParseException(f"\"settings\" key in JSON has invalid data \"{settings}\"!", context)

        # Now, parse out the settings in the dict.
        parsedsettings = [Setting.from_json(config.filename, s, [*context, f"settings[{i}]"]) for (i, s) in enumerate(settings)]

        # Finally, go through our config and match up settings to their values.
        loadedsettings = config.settings
        for setting in loadedsettings:
            for parsedsetting in parsedsettings:
                if parsedsetting.name.lower() == setting.name.lower():
                    # It's a match
                    setting.current = parsedsetting.current
                    break
            else:
                raise JSONParseException(f"Setting \"{setting.name}\" could not be found in JSON!", context)

        return Settings(config.filename, loadedsettings, type=type)

    def to_json(self) -> Dict[str, Any]:
        return {
            'type': self.type.name,
            'filename': self.filename if self.filename != NO_FILE else None,
            'settings': [
                setting.to_json() for setting in self.settings
            ],
        }

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


class SettingsWrapper:
    # A wrapper class to hold both a system and game settings section together.

    def __init__(self, serial: bytes, system: Settings, game: Settings) -> None:
        self.serial = serial
        self.system = system
        self.game = game

        self.system.type = SettingType.SYSTEM
        self.game.type = SettingType.GAME

    @staticmethod
    def from_json(settings_files: Dict[str, str], jsondict: Dict[str, Any], context: List[str]) -> "SettingsWrapper":
        # First, verify that we at least know about the system settings file.
        if "system.settings" not in settings_files:
            raise FileNotFoundError("system.settings does not seem to exist in settings definition directory!")

        # Now, grab the sections that we know we can parse.
        serial = jsondict.get('serial')
        if not isinstance(serial, str) or len(serial) != 4:
            raise JSONParseException(f"\"serial\" key in JSON has invalid data \"{serial}\"!", context)
        serialbytes = serial.encode('ascii')
        gamejson = jsondict.get('game')
        if not isinstance(gamejson, dict):
            raise JSONParseException(f"\"game\" key in JSON has invalid data \"{gamejson}\"!", context)
        systemjson = jsondict.get('system')
        if not isinstance(systemjson, dict):
            raise JSONParseException(f"\"system\" key in JSON has invalid data \"{systemjson}\"!", context)

        # First, load the system settings.
        with open(os.path.join(settings_files["system.settings"]), "r") as fp:
            systemdata = fp.read()
        systemconfig = SettingsConfig.from_data("system.settings", systemdata)

        # Now load the game settings, or if it doesn't exist, default to only allowing system settings to be set.
        gameconfig = SettingsManager._serial_to_config(settings_files, serialbytes) or SettingsConfig.blank()

        # Finally parse the EEPRom based on the config.
        system = Settings.from_json(SettingType.SYSTEM, systemconfig, systemjson, [*context, "system"])
        game = Settings.from_json(SettingType.GAME, gameconfig, gamejson, [*context, "game"])
        return SettingsWrapper(serialbytes, system, game)

    def to_json(self) -> Dict[str, Any]:
        return {
            'serial': self.serial.decode('ascii'),
            'system': self.system.to_json(),
            'game': self.game.to_json(),
        }

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


class SettingsConfig:
    # A class that can manifest a list of settings given a particular
    # file. It is not responsible for parsing any settings. It is only
    # responsible for creating the list of settings given a settings
    # definition file.

    def __init__(self, filename: str, settings: List[Setting]) -> None:
        self.filename = filename
        self.settings = settings

    @staticmethod
    def blank() -> "SettingsConfig":
        # It would be weird to display "NO FILE" to the user when there is
        # an error, but virtually all errors arise from parsing the settings
        # file itself, so if this is blank its unlikely errors will happen.
        return SettingsConfig(NO_FILE, [])

    @staticmethod
    def __get_kv(filename: str, name: str, setting: str) -> Dict[int, str]:
        realsetting = setting.strip()
        if realsetting.startswith("values are "):
            realsetting = realsetting[11:]
        if realsetting.startswith("value is "):
            realsetting = realsetting[9:]

        try:
            if "-" in realsetting:
                if " to " in realsetting:
                    raise SettingsParseException(
                        f"Setting \"{name}\" cannot have a range for valid values that includes a dash! \"{setting}\" should be specified like \"20 to E0\".",
                        filename,
                    )

                k, v = realsetting.split("-", 1)
                k = k.strip().replace(" ", "").replace("\t", "")
                key = int(k, 16)
                value = v.strip()

                return {key: value}
            else:
                if " to " in realsetting:
                    low, high = realsetting.split(" to ", 1)
                    low = low.strip()
                    high = high.strip()
                    low = low.strip().replace(" ", "").replace("\t", "")
                    high = high.strip().replace(" ", "").replace("\t", "")

                    retdict: Dict[int, str] = {}
                    for x in range(int(low, 16), int(high, 16) + 1):
                        retdict[x] = f"{x}"
                    return retdict
                else:
                    key = int(realsetting.strip(), 16)
                    value = f"{key}"

                    return {key: value}
        except ValueError:
            raise SettingsParseException(
                f"Failed to parse setting \"{name}\", could not understand value \"{setting}\".",
                filename,
            )

    @staticmethod
    def __get_vals(filename: str, name: str, setting: str) -> Tuple[str, List[int]]:
        try:
            name, rest = setting.split(" is ", 1)
            name = name.strip()
            vals: List[int] = []

            for val in rest.split(" or "):
                val = val.strip().replace(" ", "").replace("\t", "")
                vals.append(int(val, 16))

            return name, vals
        except ValueError:
            raise SettingsParseException(
                f"Failed to parse setting \"{name}\", could not understand if condition \"{setting}\".",
                filename,
            )

    @staticmethod
    def from_data(filename: str, data: str) -> "SettingsConfig":
        rawlines = data.splitlines()
        lines: List[str] = []
        settings: List[Setting] = []

        for line in rawlines:
            if not line.strip():
                # Ignore empty lines.
                continue
            if line.strip().startswith("#"):
                # Ignore comments.
                continue

            if ":" not in line:
                # Assume that this is a setting entry.
                if not lines:
                    raise SettingsParseException(f"Missing setting name before size, read-only specifier, defaults or value in \"{line}\". Perhaps you forgot a colon?", filename)

                cur = lines[-1]
                if cur.strip()[-1] == ":":
                    cur = cur + line
                else:
                    cur = cur + "," + line

                lines[-1] = cur
            else:
                # Assume that this is a full setting.
                lines.append(line)

        for line in lines:
            # First, get the name as well as the size and any restrictions.
            name, rest = line.split(":", 1)
            name = name.strip()
            rest = rest.strip()

            # Now, figure out what size it should be.
            size = SettingSizeEnum.UNKNOWN
            length = 1
            read_only: Union[bool, ReadOnlyCondition] = False
            values: Dict[int, str] = {}
            default: Optional[Union[int, DefaultConditionGroup]] = None

            if "," in rest:
                restbits = [r.strip() for r in rest.split(",")]
            else:
                restbits = [rest]

            for bit in restbits:
                if "byte" in bit or "nibble" in bit or "half-byte" in bit:
                    if " " in bit:
                        lenstr, units = bit.split(" ", 1)
                        length = int(lenstr.strip())
                        units = units.strip()
                    else:
                        units = bit.strip()

                    if "nibble" in units or "half-byte" in units:
                        size = SettingSizeEnum.NIBBLE
                    elif "byte" in units:
                        size = SettingSizeEnum.BYTE
                    else:
                        raise SettingsParseException(f"Unrecognized unit \"{units}\" for setting \"{name}\". Perhaps you misspelled \"byte\" or \"half-byte\"?", filename)
                    if size != SettingSizeEnum.BYTE and length != 1:
                        raise SettingsParseException(f"Invalid length \"{length}\" for setting \"{name}\". You should only specify a length for bytes.", filename)

                elif "read-only" in bit:
                    condstr = None
                    if " if " in bit:
                        readonlystr, condstr = bit.split(" if ", 1)
                        negate = True
                    elif " unless " in bit:
                        readonlystr, condstr = bit.split(" unless ", 1)
                        negate = False
                    else:
                        # Its unconditionally read-only.
                        read_only = True
                        readonlystr = bit

                    if readonlystr.strip() != "read-only":
                        raise SettingsParseException(f"Cannot parse read-only condition \"{bit}\" for setting \"{name}\"!", filename)
                    if condstr is not None:
                        condname, condvalues = SettingsConfig.__get_vals(filename, name, condstr)
                        read_only = ReadOnlyCondition(filename, name, condname, condvalues, negate)

                elif "default" in bit:
                    if " is " in bit:
                        defstr, rest = bit.split(" is ", 1)
                        if defstr.strip() != "default":
                            raise SettingsParseException(f"Cannot parse default \"{bit}\" for setting \"{name}\"!", filename)

                        condstr = None
                        if " if " in rest:
                            rest, condstr = rest.split(" if ", 1)
                            negate = False
                        elif " unless " in rest:
                            rest, condstr = rest.split(" unless ", 1)
                            negate = True
                        else:
                            # Its unconditionally a default.
                            pass

                        rest = rest.strip().replace(" ", "").replace("\t", "")
                        defbytes = bytes([int(rest[i:(i + 2)], 16) for i in range(0, len(rest), 2)])
                        if size != SettingSizeEnum.UNKNOWN and len(defbytes) == 1:
                            defaultint = defbytes[0]
                        else:
                            if size == SettingSizeEnum.NIBBLE:
                                defaultint = struct.unpack("<B", defbytes[0:1])[0]
                            elif size == SettingSizeEnum.BYTE:
                                if length == 1:
                                    defaultint = struct.unpack("<B", defbytes[0:1])[0]
                                elif length == 2:
                                    defaultint = struct.unpack("<H", defbytes[0:2])[0]
                                elif length == 4:
                                    defaultint = struct.unpack("<I", defbytes[0:4])[0]
                                else:
                                    raise SettingsParseException(
                                        f"Cannot convert default \"{bit}\" for setting \"{name}\" because we don't know how to handle length \"{length}\"!",
                                        filename,
                                    )
                            else:
                                raise SettingsParseException(f"Must place default \"{bit}\" after size specifier in setting \"{name}\"!", filename)

                        if condstr is None:
                            if default is not None:
                                if isinstance(default, DefaultConditionGroup):
                                    raise SettingsParseException(f"Cannot specify an unconditional default alongside conditional defaults for setting \"{name}\"!", filename)
                                else:
                                    raise SettingsParseException(f"Cannot specify more than one default for setting \"{name}\"!", filename)
                            default = defaultint
                        else:
                            if default is None:
                                default = DefaultConditionGroup(filename, name, [])
                            if not isinstance(default, DefaultConditionGroup):
                                raise SettingsParseException(f"Cannot specify an unconditional default alongside conditional defaults for setting \"{name}\"!", filename)

                            condname, condvalues = SettingsConfig.__get_vals(filename, name, condstr)
                            default.conditions.append(DefaultCondition(condname, condvalues, negate, defaultint))
                    else:
                        raise SettingsParseException(f"Cannot parse default for setting \"{name}\"! Specify defaults like \"default is 0\".", filename)

                else:
                    # Assume this is a setting value.
                    values.update(SettingsConfig.__get_kv(filename, name, bit))

            if size == SettingSizeEnum.UNKNOWN:
                raise SettingsParseException(f"Setting \"{name}\" is missing a size specifier!", filename)
            if read_only is not True and not values:
                raise SettingsParseException(f"Setting \"{name}\" is missing any valid values!", filename)

            settings.append(
                Setting(
                    name,
                    size,
                    length,
                    read_only,
                    values,
                    default=default,
                )
            )

        # Verify that nibbles come in pairs.
        halves = 0
        for setting in settings:
            if setting.size == SettingSizeEnum.NIBBLE:
                halves = 1 - halves
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsParseException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", filename)

        return SettingsConfig(filename, settings)

    def defaults(self) -> bytes:
        pending = 0
        halves = 0
        defaults: List[bytes] = []

        for setting in self.settings:
            if setting.default is None:
                default = 0
            elif isinstance(setting.default, int):
                default = setting.default
            elif isinstance(setting.default, DefaultConditionGroup):
                # Must evaluate settings to figure out the default for this.
                default = setting.default.evaluate(self.filename, setting.name, self.settings)

            if setting.size == SettingSizeEnum.NIBBLE:
                if halves == 0:
                    pending = (default & 0xF) << 4
                else:
                    defaults.append(bytes([(default & 0xF) | pending]))

                if halves == 0:
                    halves = 1
                else:
                    halves = 0
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", self.filename)
                if setting.length == 1:
                    defaults.append(struct.pack("<B", default))
                elif setting.length == 2:
                    defaults.append(struct.pack("<H", default))
                elif setting.length == 4:
                    defaults.append(struct.pack("<I", default))
                else:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", self.filename)

        return b"".join(defaults)


def get_default_settings_directory() -> str:
    # Specifically for projects including this code as a 3rd-party dependency,
    # look up where we stick the default settings definitions files and return
    # that path as a string, suitable for passing into the "directory" param of
    # SettingsManager.
    return os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "definitions"))


class SettingsManager:
    # A manager class that can handle manifesting and saving settings given a directory
    # of definition files.

    def __init__(self, directory: str) -> None:
        self.__directory = directory

    @property
    def files(self) -> Dict[str, str]:
        return {
            f: os.path.join(self.__directory, f)
            for f in os.listdir(self.__directory)
            if os.path.isfile(os.path.join(self.__directory, f)) and f.endswith(".settings")
        }

    @staticmethod
    def _serial_to_config(files: Dict[str, str], serial: bytes) -> Optional[SettingsConfig]:
        fname = f"{serial.decode('ascii')}.settings"

        if fname not in files:
            return None

        with open(files[fname], "r") as fp:
            data = fp.read()

        return SettingsConfig.from_data(fname, data)

    def from_serial(self, serial: bytes) -> SettingsWrapper:
        config = self._serial_to_config(self.files, serial)
        defaults = None
        if config is not None:
            defaults = config.defaults()

        return self.from_eeprom(NaomiEEPRom.default(serial, game_defaults=defaults).data)

    def from_eeprom(self, data: bytes) -> SettingsWrapper:
        # First grab the parsed EEPRom so we can get the serial.
        eeprom = NaomiEEPRom(data)

        # Now load the system settings.
        with open(os.path.join(self.__directory, "system.settings"), "r") as fp:
            systemdata = fp.read()
        systemconfig = SettingsConfig.from_data("system.settings", systemdata)

        # Now load the game settings, or if it doesn't exist, default to only
        # allowing system settings to be set.
        gameconfig = self._serial_to_config(self.files, eeprom.serial) or SettingsConfig.blank()

        # Finally parse the EEPRom based on the config.
        system = Settings.from_config(SettingType.SYSTEM, systemconfig, eeprom)
        game = Settings.from_config(SettingType.GAME, gameconfig, eeprom)
        return SettingsWrapper(eeprom.serial, system, game)

    def from_json(self, jsondict: Dict[str, Any], context: Optional[List[str]] = None) -> SettingsWrapper:
        return SettingsWrapper.from_json(self.files, jsondict, context or [])

    def to_eeprom(self, settings: SettingsWrapper) -> bytes:
        # First, create the EEPROM.
        eeprom = NaomiEEPRom.default(settings.serial)

        # Now, update the game length.
        eeprom.length = settings.game.length

        for section, settingsgroup in [
            (eeprom.system, settings.system),
            (eeprom.game, settings.game),
        ]:
            pending = 0
            halves = 0
            location = 0

            if not section.valid:
                # If we couldn't make this section correct, completely skip out on it.
                continue

            for setting in settingsgroup.settings:
                # First, calculate what the default should be in case we need to use it.
                if setting.default is None:
                    default = None
                elif isinstance(setting.default, int):
                    default = setting.default
                elif isinstance(setting.default, DefaultConditionGroup):
                    # Must evaluate settings to figure out the default for this.
                    default = setting.default.evaluate(settingsgroup.filename, setting.name, settingsgroup.settings)

                # Now, figure out if we should defer to the default over the current value
                # (if it is read-only) or if we should use the current value.
                if isinstance(setting.read_only, ReadOnlyCondition):
                    read_only = setting.read_only.evaluate(settingsgroup.settings)
                elif setting.read_only is True:
                    read_only = True
                elif setting.read_only is False:
                    read_only = False

                if read_only:
                    # If it is read-only, then only take the current value if the default doesn't
                    # exist. This lets settings that are selectively read-only get a conditional
                    # default if one exists that takes precedence over the current value.
                    value = setting.current if default is None else default
                else:
                    # If the setting is not read-only, then only take the default if the current
                    # value is None.
                    value = default if setting.current is None else setting.current

                # Now, write out the setting by updating the EEPROM in the correct location.
                if setting.size == SettingSizeEnum.NIBBLE:
                    # First, if we have anything to write, write it.
                    if value is not None:
                        if halves == 0:
                            pending = (value & 0xF) << 4
                        else:
                            section[location] = (value & 0xF) | pending

                    # Now, update our position.
                    if halves == 0:
                        halves = 1
                    else:
                        halves = 0
                        location += 1

                elif setting.size == SettingSizeEnum.BYTE:
                    # First, make sure we aren't in a pending nibble state.
                    if halves != 0:
                        raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", settingsgroup.filename)

                    if setting.length not in {1, 2, 4}:
                        raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", settingsgroup.filename)

                    if value is not None:
                        if setting.length == 1:
                            section[location:(location + 1)] = struct.pack("<B", value)
                        elif setting.length == 2:
                            section[location:(location + 2)] = struct.pack("<H", value)
                        elif setting.length == 4:
                            section[location:(location + 4)] = struct.pack("<I", value)

                    # Now, update our position.
                    location += setting.length

        return eeprom.data



#! /usr/bin/env python3
import os
import struct
from typing import Optional, Tuple

from naomi import NaomiRom, NaomiRomSection, NaomiEEPRom


def get_default_trojan() -> bytes:
    # Specifically for projects including this code as a 3rd-party dependency,
    # look up where we stick the default already-compiled trojan and return it
    # as bytes that can be passed into the "trojan" param of the
    # NaomiSettingsPatcher constructor. This way, they don't need to know our
    # internal directory layout.
    root = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))
    trojan = os.path.join(root, 'homebrew', 'settingstrojan', 'settingstrojan.bin')
    with open(trojan, "rb") as bfp:
        return bfp.read()


class NaomiSettingsPatcherException(Exception):
    pass


class NaomiSettingsDate:
    def __init__(self, year: int, month: int, day: int) -> None:
        self.year = year
        self.month = month
        self.day = day


class NaomiSettingsInfo:
    def __init__(self, sentinel: bool, debug: bool, date: Tuple[int, int, int]) -> None:
        self.enable_sentinel = sentinel
        self.enable_debugging = debug
        self.date = NaomiSettingsDate(date[0], date[1], date[2])


class NaomiSettingsPatcher:
    def __init__(self, rom: bytes, trojan: bytes) -> None:
        self.data = rom
        self.__trojan = trojan

    @staticmethod
    def __change(binfile: bytes, tochange: bytes, loc: int) -> bytes:
        return binfile[:loc] + tochange + binfile[(loc + len(tochange)):]

    @staticmethod
    def __patch_bytesequence(data: bytes, sentinel: int, replacement: bytes) -> bytes:
        length = len(replacement)
        for i in range(len(data) - length + 1):
            if all(x == sentinel for x in data[i:(i + length)]):
                return NaomiSettingsPatcher.__change(data, replacement, i)

        raise NaomiSettingsPatcherException("Couldn't find spot to patch in data!")

    @staticmethod
    def __get_config(data: bytes) -> Tuple[int, int, bool, bool, Tuple[int, int, int]]:
        # Returns a tuple consisting of the original EXE start address and
        # the desired trojan start address, whether sentinel mode is enabled
        # and whether debug printing is enabled, and the date string of the
        # trojan we're using.
        for i in range(len(data) - 28):
            if all(x == 0xEE for x in data[i:(i + 4)]) and all(x == 0xEE for x in data[(i + 24):(i + 28)]):
                original_start, trojan_start, sentinel, debug, date = struct.unpack("<IIIII", data[(i + 4):(i + 24)])
                if sentinel not in {0, 1, 0xCFCFCFCF} or debug not in {0, 1, 0xDDDDDDDD}:
                    continue

                day = date % 100
                month = (date // 100) % 100
                year = (date // 10000)

                return (
                    original_start,
                    trojan_start,
                    sentinel != 0,
                    debug != 0,
                    (year, month, day),
                )

        raise NaomiSettingsPatcherException("Couldn't find config in executable!")

    def get_serial(self) -> bytes:
        # Parse the ROM header so we can grab the game serial code.
        naomi = NaomiRom(self.data)
        return naomi.serial

    def get_info(self) -> Optional[NaomiSettingsInfo]:
        # Parse the ROM header so we can narrow our search.
        naomi = NaomiRom(self.data)

        # Only look at main executables.
        executable = naomi.main_executable
        for sec in executable.sections:
            # Constrain the search to the section that we jump to, since that will always
            # be where our trojan is.
            if executable.entrypoint >= sec.load_address and executable.entrypoint < (sec.load_address + sec.length):
                try:
                    # Grab the old entrypoint from the existing modification since the ROM header
                    # entrypoint will be the old trojan EXE.
                    data = self.data[sec.offset:(sec.offset + sec.length)]
                    _, _, sentinel, debug, date = self.__get_config(data)

                    return NaomiSettingsInfo(sentinel, debug, date)
                except Exception:
                    continue

        return None

    def get_settings(self) -> Optional[bytes]:
        # Parse the ROM header so we can narrow our search.
        naomi = NaomiRom(self.data)

        # Only look at main executables.
        executable = naomi.main_executable
        for sec in executable.sections:
            # Constrain the search to the section that we jump to, since that will always
            # be where our trojan is.
            if executable.entrypoint >= sec.load_address and executable.entrypoint < (sec.load_address + sec.length):
                try:
                    # Grab the old entrypoint from the existing modification since the ROM header
                    # entrypoint will be the old trojan EXE.
                    data = self.data[sec.offset:(sec.offset + sec.length)]
                    self.__get_config(data)

                    # Returns the requested EEPRom settings that should be written prior
                    # to the game starting.
                    for i in range(len(data) - 128):
                        if NaomiEEPRom.validate(data[i:(i + 128)]):
                            return data[i:(i + 128)]
                except Exception:
                    pass

        # Couldn't find a section that matched.
        return None

    def put_settings(self, settings: bytes, *, enable_sentinel: bool = False, enable_debugging: bool = False, verbose: bool = False) -> None:
        # First, parse the ROM we were given.
        data = self.data
        naomi = NaomiRom(data)

        # First, we need to modify the settings trojan with this ROM's load address and
        # the EEPROM we want to add. Make sure the EEPRom we were given is valid.
        if len(settings) != 128:
            raise NaomiSettingsPatcherException("Invalid length of settings!")
        if not NaomiEEPRom.validate(settings):
            raise NaomiSettingsPatcherException("Settings is incorrectly formed!")
        if naomi.serial != settings[3:7] or naomi.serial != settings[21:25]:
            raise NaomiSettingsPatcherException("Settings is not for this game!")

        # Now we need to add an EXE init section to the ROM.
        exe = self.__trojan[:]
        executable = naomi.main_executable
        _, location, _, _, _ = self.__get_config(exe)

        for sec in executable.sections:
            if sec.load_address == location:
                # Grab the old entrypoint from the existing modification since the ROM header
                # entrypoint will be the old trojan EXE.
                entrypoint, _, _, _, _ = self.__get_config(data[sec.offset:(sec.offset + sec.length)])
                exe = self.__patch_bytesequence(exe, 0xAA, struct.pack("<I", entrypoint))
                exe = self.__patch_bytesequence(exe, 0xBB, settings)
                exe = self.__patch_bytesequence(exe, 0xCF, struct.pack("<I", 1 if enable_sentinel else 0))
                exe = self.__patch_bytesequence(exe, 0xDD, struct.pack("<I", 1 if enable_debugging else 0))

                # We can reuse this section, but first we need to get rid of the old patch.
                if sec.offset + sec.length == len(data):
                    # We can just stick the new file right on top of where the old was.
                    if verbose:
                        print("Overwriting old settings in existing ROM section.")

                    # Cut off the old section, add our new section, make sure the length is correct.
                    data = data[:sec.offset] + exe
                    sec.length = len(exe)
                else:
                    # It is somewhere in the middle of an executable, zero it out and
                    # then add this section to the end of the ROM.
                    if verbose:
                        print("Zeroing out old settings in existing ROM section and attaching new settings to the end of the file.")

                    # Patch the executable with the correct settings and entrypoint.
                    data = self.__change(data, b"\0" * sec.length, sec.offset)
                    sec.offset = len(data)
                    sec.length = len(exe)
                    sec.load_address = location
                    data += exe
                break
        else:
            if len(executable.sections) >= 8:
                raise NaomiSettingsPatcherException("ROM already has the maximum number of init sections!")

            # Add a new section to the end of the rom for this binary data.
            if verbose:
                print("Attaching settings to a new ROM section at the end of the file.")

            executable.sections.append(
                NaomiRomSection(
                    offset=len(data),
                    load_address=location,
                    length=len(exe),
                )
            )

            # Patch the executable with the correct settings and entrypoint.
            exe = self.__patch_bytesequence(exe, 0xAA, struct.pack("<I", executable.entrypoint))
            exe = self.__patch_bytesequence(exe, 0xBB, settings)
            exe = self.__patch_bytesequence(exe, 0xCF, struct.pack("<I", 1 if enable_sentinel else 0))
            exe = self.__patch_bytesequence(exe, 0xDD, struct.pack("<I", 1 if enable_debugging else 0))
            data += exe

        executable.entrypoint = location

        # Generate new header and attach executable to end of data.
        naomi.main_executable = executable

        # Now, save this back so it can be read by other callers.
        self.data = naomi.data + data[naomi.HEADER_LENGTH:]



from naomi.eeprom import NaomiEEPRom, NaomiEEPRomException
from naomi.generic_patch import force_freeplay, force_no_attract_sound
from naomi.rom import NaomiRom, NaomiExecutable, NaomiRomSection, NaomiRomException
from naomi.settings_patch import NaomiSettingsPatcher, NaomiSettingsPatcherException, get_default_trojan

__all__ = [
    "NaomiRom",
    "NaomiExecutable",
    "NaomiRomSection",
    "NaomiRomException",
    "NaomiEEPRom",
    "NaomiEEPRomException",
    "NaomiSettingsPatcher",
    "NaomiSettingsPatcherException",
    "force_freeplay",
    "force_no_attract_sound",
    "get_default_trojan",
]



import datetime
import struct
from typing import List, cast


class NaomiRomException(Exception):
    pass


class NaomiRomSection:
    def __init__(self, offset: int, length: int, load_address: int) -> None:
        self.offset = offset
        self.length = length
        self.load_address = load_address

    def __repr__(self) -> str:
        return f"NaomiRomSection(offset={self.offset}, length={self.length}, load_address={hex(self.load_address)})"


class NaomiExecutable:
    def __init__(self, entrypoint: int, sections: List[NaomiRomSection]) -> None:
        self.sections = sections
        self.entrypoint = entrypoint

    def __repr__(self) -> str:
        return f"NaomiExecutable(entrypoint={hex(self.entrypoint)}, sections={repr(self.sections)})"


class NaomiEEPROMDefaults:
    def __init__(
        self,
        region: int,
        apply_settings: bool,
        force_vertical: bool,
        force_silent: bool,
        chute: str,
        coin_setting: int,
        coin_1_rate: int,
        coin_2_rate: int,
        credit_rate: int,
        bonus: int,
        sequences: List[str],
    ):
        self.region = region
        self.apply_settings = apply_settings
        self.force_vertical = force_vertical
        self.force_silent = force_silent
        self.chute = chute
        self.coin_setting = coin_setting
        self.coin_1_rate = coin_1_rate
        self.coin_2_rate = coin_2_rate
        self.credit_rate = credit_rate
        self.bonus = bonus
        self.sequences = sequences

    def __repr__(self) -> str:
        return (
            "NaomiEEPROMDefaults("
            f"region={repr(self.region)}, "
            f"apply_settings={repr(self.apply_settings)}, "
            f"force_vertical={repr(self.force_vertical)}, "
            f"force_silent={repr(self.force_silent)}, "
            f"chute={repr(self.chute)}, "
            f"coin_setting={repr(self.coin_setting)}, "
            f"coin_1_rate={repr(self.coin_1_rate)}, "
            f"coin_2_rate={repr(self.coin_2_rate)}, "
            f"credit_rate={repr(self.credit_rate)}, "
            f"bonus={repr(self.bonus)}, "
            f"sequences={repr(self.sequences)}"
            ")"
        )


class NaomiRom:

    REGION_JAPAN = 0
    REGION_USA = 1
    REGION_EXPORT = 2
    REGION_KOREA = 3
    REGION_AUSTRALIA = 4

    HEADER_LENGTH = 0x500

    def __init__(self, data: bytes) -> None:
        self.data = data[:NaomiRom.HEADER_LENGTH]
        self.data = self.data + (b'\0' * (NaomiRom.HEADER_LENGTH - len(self.data)))

    @staticmethod
    def default() -> "NaomiRom":
        return NaomiRom(b'NAOMI           ' + (b'\0' * (NaomiRom.HEADER_LENGTH - 17)) + b'\xFF')

    @property
    def valid(self) -> bool:
        return self.data[0x000:0x010] == b'NAOMI           '

    def _raise_on_invalid(self) -> None:
        if not self.valid:
            raise NaomiRomException("Not a valid Naomi ROM!")

    def _inject(self, offset: int, data: bytes) -> None:
        self.data = self.data[:offset] + data + self.data[(offset + len(data)):]
        if len(self.data) != self.HEADER_LENGTH:
            raise Exception("Logic error!")

    def _inject_str(self, offset: int, string: str, pad_length: int) -> None:
        data = string.encode('ascii')
        data = data[:pad_length]
        data = data + (b' ' * (pad_length - len(data)))
        self._inject(offset, data)

    def _inject_uint32(self, offset: int, value: int) -> None:
        self._inject(offset, struct.pack("<I", value))

    def _inject_uint16(self, offset: int, value: int) -> None:
        self._inject(offset, struct.pack("<H", value))

    def _inject_uint8(self, offset: int, value: int) -> None:
        self._inject(offset, struct.pack("<B", value))

    def _sanitize_str(self, data: bytes) -> str:
        return data.decode('ascii').replace("\x00", " ").strip()

    def _sanitize_uint32(self, offset: int) -> int:
        return cast(int, struct.unpack("<I", self.data[offset:(offset + 4)])[0])

    def _sanitize_uint16(self, offset: int) -> int:
        return cast(int, struct.unpack("<H", self.data[offset:(offset + 2)])[0])

    def _sanitize_uint8(self, offset: int) -> int:
        return cast(int, struct.unpack("<B", self.data[offset:(offset + 1)])[0])

    @property
    def publisher(self) -> str:
        self._raise_on_invalid()
        return self._sanitize_str(self.data[0x010:0x030])

    @publisher.setter
    def publisher(self, val: str) -> None:
        self._raise_on_invalid()
        self._inject_str(0x010, val, 0x030 - 0x010)

    @property
    def names(self) -> List[str]:
        self._raise_on_invalid()

        return [
            self._sanitize_str(self.data[0x030:0x050]),
            self._sanitize_str(self.data[0x050:0x070]),
            self._sanitize_str(self.data[0x070:0x090]),
            self._sanitize_str(self.data[0x090:0x0B0]),
            self._sanitize_str(self.data[0x0B0:0x0D0]),
        ]

    @names.setter
    def names(self, val: List[str]) -> None:
        self._raise_on_invalid()
        if len(val) != 5:
            raise NaomiRomException("Expected list of five strings for names!")
        val.extend(["", "", ""])

        for i, (start, end) in enumerate([
            (0x030, 0x050),
            (0x050, 0x070),
            (0x070, 0x090),
            (0x090, 0x0B0),
            (0x0B0, 0x0D0),
            # The following are regions that we don't care to write since they
            # are not used in practice.
            # (0x0D0, 0x0F0),
            # (0x0F0, 0x110),
            # (0x110, 0x130),
        ]):
            self._inject_str(start, val[i], end - start)

    @property
    def sequencetexts(self) -> List[str]:
        self._raise_on_invalid()

        return [
            self._sanitize_str(self.data[0x260:0x280]),
            self._sanitize_str(self.data[0x280:0x2A0]),
            self._sanitize_str(self.data[0x2A0:0x2C0]),
            self._sanitize_str(self.data[0x2C0:0x2E0]),
            self._sanitize_str(self.data[0x2E0:0x300]),
            self._sanitize_str(self.data[0x300:0x320]),
            self._sanitize_str(self.data[0x320:0x340]),
            self._sanitize_str(self.data[0x340:0x360]),
        ]

    @sequencetexts.setter
    def sequencetexts(self, val: List[str]) -> None:
        self._raise_on_invalid()
        if len(val) > 8:
            raise NaomiRomException("Expected list of eight strings for sequence texts!")
        while len(val) < 8:
            val.append('')

        for i, (start, end) in enumerate([
            (0x260, 0x280),
            (0x280, 0x2A0),
            (0x2A0, 0x2C0),
            (0x2C0, 0x2E0),
            (0x2E0, 0x300),
            (0x300, 0x320),
            (0x320, 0x340),
            (0x340, 0x360),
        ]):
            self._inject_str(start, val[i], end - start)

    @property
    def defaults(self) -> List[NaomiEEPROMDefaults]:
        self._raise_on_invalid()

        sections: List[NaomiEEPROMDefaults] = []
        texts: List[str] = self.sequencetexts

        # The following list must remain sorted in order to output correctly.
        for offset in [self.REGION_JAPAN, self.REGION_USA, self.REGION_EXPORT, self.REGION_KOREA, self.REGION_AUSTRALIA]:
            location = 0x1E0 + (0x10 * offset)

            sections.append(NaomiEEPROMDefaults(
                region=offset,
                apply_settings=self._sanitize_uint8(location + 0) == 1,
                force_vertical=(self._sanitize_uint8(location + 1) & 0x1) != 0,
                force_silent=(self._sanitize_uint8(location + 1) & 0x2) != 0,
                chute="individual" if self._sanitize_uint8(location + 2) != 0 else "common",
                coin_setting=self._sanitize_uint8(location + 3),
                coin_1_rate=self._sanitize_uint8(location + 4),
                coin_2_rate=self._sanitize_uint8(location + 5),
                credit_rate=self._sanitize_uint8(location + 6),
                bonus=self._sanitize_uint8(location + 7),
                sequences=[texts[self._sanitize_uint8(location + 8 + x)] for x in range(8)],
            ))
        return sections

    @defaults.setter
    def defaults(self, val: List[NaomiEEPROMDefaults]) -> None:
        self._raise_on_invalid()

        # Go through the list, back-converting each structure
        for defaults in val:
            if defaults.region not in {
                self.REGION_JAPAN, self.REGION_USA, self.REGION_EXPORT, self.REGION_KOREA, self.REGION_AUSTRALIA
            }:
                raise NaomiRomException(f"Invalid region {defaults.region} in defaults!")
            offset = 0x1E0 + (0x10 * defaults.region)

            self._inject_uint8(offset + 0, 1 if defaults.apply_settings else 0)

            mask: int = 0
            if defaults.force_vertical:
                mask |= 0x1
            if defaults.force_silent:
                mask |= 0x2
            self._inject_uint8(offset + 1, mask)

            if defaults.chute == "individual":
                self._inject_uint8(offset + 2, 1)
            elif defaults.chute == "common":
                self._inject_uint8(offset + 2, 0)
            else:
                raise NaomiRomException(f"Invalid chute type {defaults.chute} in defaults!")

            self._inject_uint8(offset + 3, defaults.coin_setting)
            self._inject_uint8(offset + 4, defaults.coin_1_rate)
            self._inject_uint8(offset + 5, defaults.coin_2_rate)
            self._inject_uint8(offset + 6, defaults.credit_rate)
            self._inject_uint8(offset + 7, defaults.bonus)

            # Calculate sequence offsets
            sequences: List[str] = list(defaults.sequences)
            if len(sequences) > 8:
                raise NaomiRomException("Invalid number of sequence texts for defaults, expected at most 8!")

            # First, default them all to 0
            self._inject(offset + 8, b'\0' * 8)

            # Now, attempt to insert texts that are non-default
            for i, text in enumerate(sequences):
                # First, if it is empty, default it
                if text == "":
                    continue

                # Try to find this sequence text in our global list.
                # This is not cached because it can change.
                validtexts = self.sequencetexts
                for off, existingtext in enumerate(validtexts):
                    if text == existingtext:
                        self._inject_uint8(offset + 8 + i, off)
                        break
                else:
                    # We didn't find one, add it to the end!
                    for off, existingtext in enumerate(validtexts):
                        if existingtext == "":
                            validtexts[off] = text
                            self.sequencetexts = validtexts
                            self._inject_uint8(offset + 8 + i, off)
                            break
                    else:
                        # We didn't find room!
                        raise NaomiRomException(f"Not enough room in sequence text table for {text}!")

    @property
    def date(self) -> datetime.date:
        self._raise_on_invalid()
        year = self._sanitize_uint16(0x130)
        month = self._sanitize_uint8(0x132)
        day = self._sanitize_uint8(0x133)
        if year == 0:
            year = 2000
        if month == 0:
            month = 1
        if day == 0:
            day = 1
        return datetime.date(year, month, day)

    @date.setter
    def date(self, date: datetime.date) -> None:
        self._raise_on_invalid()
        self._inject_uint16(0x130, date.year)
        self._inject_uint8(0x132, date.month)
        self._inject_uint8(0x133, date.day)

    @property
    def serial(self) -> bytes:
        self._raise_on_invalid()
        return self.data[0x134:0x138]

    @serial.setter
    def serial(self, serial: bytes) -> None:
        self._raise_on_invalid()
        if len(serial) != 4:
            raise NaomiRomException("Serial length should be 4 bytes!")
        if serial[0:1] != b'B':
            raise NaomiRomException("Serial should start with B!")
        self._inject(0x134, serial)

    @property
    def regions(self) -> List[int]:
        self._raise_on_invalid()
        mask = self._sanitize_uint8(0x428)
        regions: List[int] = []
        for offset in [self.REGION_JAPAN, self.REGION_USA, self.REGION_EXPORT, self.REGION_KOREA, self.REGION_AUSTRALIA]:
            if ((mask >> offset) & 0x1) != 0:
                regions.append(offset)
        return regions

    @regions.setter
    def regions(self, regions: List[int]) -> None:
        self._raise_on_invalid()

        mask: int = 0
        for region in regions:
            if region not in {self.REGION_JAPAN, self.REGION_USA, self.REGION_EXPORT, self.REGION_KOREA, self.REGION_AUSTRALIA}:
                raise NaomiRomException(f"Region {region} not recognized!")
            mask |= 1 << region
        self._inject_uint8(0x428, mask)

    @property
    def players(self) -> List[int]:
        self._raise_on_invalid()
        mask = self._sanitize_uint8(0x429)
        if mask == 0:
            return [1, 2, 3, 4]
        return sorted([
            (x + 1) for x in range(4)
            if ((mask >> x) & 0x1) != 0
        ])

    @players.setter
    def players(self, players: List[int]) -> None:
        self._raise_on_invalid()

        mask: int = 0
        for player in players:
            if player not in {1, 2, 3, 4}:
                raise NaomiRomException(f"Number of players {player} not valid!")
            mask |= 1 << (player - 1)
        self._inject_uint8(0x429, mask)

    @property
    def frequencies(self) -> List[int]:
        self._raise_on_invalid()
        mask = self._sanitize_uint8(0x42A)
        if mask == 0:
            return [15, 31]
        lut = [31, 15]
        return sorted([
            lut[x] for x in range(2)
            if ((mask >> x) & 0x1) != 0
        ])

    @frequencies.setter
    def frequencies(self, frequencies: List[int]) -> None:
        self._raise_on_invalid()

        mask: int = 0
        for frequency in frequencies:
            if frequency not in {15, 31}:
                raise NaomiRomException(f"Monitor frequency {frequency} not valid!")
            if frequency == 15:
                mask |= 0x2
            if frequency == 31:
                mask |= 0x1
        self._inject_uint8(0x42A, mask)

    @property
    def orientations(self) -> List[str]:
        self._raise_on_invalid()
        mask = self._sanitize_uint8(0x42B)
        lut = ['horizontal', 'vertical']
        if mask == 0:
            return lut
        return sorted([
            lut[x] for x in range(2)
            if ((mask >> x) & 0x1) != 0
        ])

    @orientations.setter
    def orientations(self, orientations: List[str]) -> None:
        self._raise_on_invalid()

        mask: int = 0
        for orientation in orientations:
            if orientation not in {'horizontal', 'vertical'}:
                raise NaomiRomException(f"Monitor orientation {orientation} not valid!")
            if orientation == "horizontal":
                mask |= 0x1
            if orientation == "vertical":
                mask |= 0x2
        self._inject_uint8(0x42B, mask)

    @property
    def servicetype(self) -> str:
        self._raise_on_invalid()
        return 'individual' if self._sanitize_uint8(0x42D) != 0 else 'common'

    @servicetype.setter
    def servicetype(self, servicetype: str) -> None:
        self._raise_on_invalid()
        if servicetype == "individual":
            self._inject_uint8(0x42D, 1)
        elif servicetype == "common":
            self._inject_uint8(0x42D, 0)
        else:
            raise NaomiRomException(f"Service type {servicetype} not valid!")

    def _get_sections(self, startoffset: int) -> List[NaomiRomSection]:
        sections: List[NaomiRomSection] = []
        for entry in range(8):
            location = startoffset + 12 * entry

            offset = self._sanitize_uint32(location)
            if offset == 0xFFFFFFFF:
                break
            sections.append(
                NaomiRomSection(
                    offset=offset,
                    length=self._sanitize_uint32(location + 8),
                    load_address=self._sanitize_uint32(location + 4),
                )
            )
        return sections

    def _put_sections(self, offset: int, sections: List[NaomiRomSection]) -> None:
        if len(sections) > 8:
            raise NaomiRomException("Cannot have more than 8 load sections in an executable!")

        for section in sections:
            self._inject_uint32(offset, section.offset)
            self._inject_uint32(offset + 4, section.load_address)
            self._inject_uint32(offset + 8, section.length)
            offset += 12

        if len(sections) < 8:
            # Put an end of list marker so we don't load any garbage data.
            self._inject_uint32(offset, 0xFFFFFFFF)

    @property
    def main_executable(self) -> NaomiExecutable:
        self._raise_on_invalid()
        return NaomiExecutable(
            sections=self._get_sections(0x360),
            entrypoint=self._sanitize_uint32(0x420),
        )

    @main_executable.setter
    def main_executable(self, val: NaomiExecutable) -> None:
        self._raise_on_invalid()
        self._put_sections(0x360, val.sections)
        self._inject_uint32(0x420, val.entrypoint)

    @property
    def test_executable(self) -> NaomiExecutable:
        self._raise_on_invalid()
        return NaomiExecutable(
            sections=self._get_sections(0x3C0),
            entrypoint=self._sanitize_uint32(0x424),
        )

    @test_executable.setter
    def test_executable(self, val: NaomiExecutable) -> None:
        self._raise_on_invalid()
        self._put_sections(0x3C0, val.sections)
        self._inject_uint32(0x424, val.entrypoint)



#! /usr/bin/env python3
from naomi.rom import NaomiRom


def _patch_rom(data: bytes, search: bytes, replace: bytes) -> bytes:
    rom = NaomiRom(data)
    if not rom.valid:
        raise Exception("ROM file does not appear to be a Naomi netboot ROM!")

    # Now, find the location of our magic string in the main executable.
    exec_data = rom.main_executable
    patch_location = None
    search_len = len(search)

    # Look through all copied sections.
    for section in exec_data.sections:
        for rloc in range(section.length - search_len):
            loc = rloc + section.offset
            if data[loc:(loc + search_len)] == search:
                patch_location = loc
                break
        if patch_location is not None:
            break

    if patch_location is None:
        raise Exception("ROM file does not have a suitable spot for the patch!")

    # Now, generate a patch with this updated data overlaid on the original rom
    return data[:patch_location] + replace + data[(patch_location + len(replace)):]


def force_freeplay(data: bytes) -> bytes:
    return _patch_rom(data, bytes([0x42, 0x84, 0xEC, 0x31, 0x0C, 0x60, 0x04, 0x1E, 0x43, 0x84]), bytes([0x1A, 0xE0]))


def force_no_attract_sound(data: bytes) -> bytes:
    return _patch_rom(data, bytes([0x40, 0x63, 0x12, 0xe2, 0xec, 0x32, 0x3c, 0x63, 0x09, 0x43]), bytes([0x00, 0xe3]))



from naomi.settings.settings import (
    SettingsParseException,
    SettingsSaveException,
    JSONParseException,
    ReadOnlyCondition,
    DefaultCondition,
    DefaultConditionGroup,
    Setting,
    Settings,
    SettingsWrapper,
    SettingsManager,
    get_default_settings_directory,
)
try:
    from naomi.settings.editor import SettingsEditor
except ModuleNotFoundError:
    # DragonCurses not installed, provide a dummy stub.
    class SettingsEditor:  # type: ignore
        def __init__(self, settings: SettingsWrapper, enable_unicode: bool = True) -> None:
            self.settings = settings

        def run(self) -> bool:
            return False


__all__ = [
    "SettingsParseException",
    "SettingsSaveException",
    "JSONParseException",
    "ReadOnlyCondition",
    "DefaultCondition",
    "DefaultConditionGroup",
    "Setting",
    "Settings",
    "SettingsWrapper",
    "SettingsManager",
    "SettingsEditor",
    "get_default_settings_directory",
]



import os
from setuptools import setup


if 'FULL_INSTALLATION' in os.environ:
    # We want to install this entire repo as an installation, so that we can
    # use it to run a netboot server.
    with open("MANIFEST.in", "w") as wfp:
        with open("MANIFEST.install", "r") as rfp:
            wfp.write(rfp.read())

    def requires(req: str) -> str:
        if "git+" not in req:
            return req
        if "#egg" not in req:
            raise Exception(f"Unknown egg package for {req}!")
        _, egg = req.split("#egg", 1)
        egg = egg.strip()
        if egg.startswith("="):
            egg = egg[1:].strip()
        return egg

    setup(
        name='netboot',
        version='0.1',
        description='Code and utilities for netbooting a Naomi/Triforce/Chihiro, including a full web and netboot server.',
        author='DragonMinded',
        license='Public Domain',
        packages=[
            # Core packages.
            'naomi',
            'naomi.settings',
            # Web server package.
            'netboot',
            'netboot.web',
        ],
        install_requires=[
            requires(req) for req in open('requirements.txt').read().split('\n') if len(req) > 0
        ],
        include_package_data=True,
        zip_safe=False,
    )
else:
    # We want to install only the parts of this repo useful as a third-party
    # package so that other code can depend on us.
    with open(os.path.join("naomi", "README.md"), "r", encoding="utf-8") as fh:
        long_description = fh.read()
    long_description += os.linesep
    with open(os.path.join("naomi", "settings", "README.md"), "r", encoding="utf-8") as fh:
        long_description += fh.read()
    long_description += os.linesep
    with open(os.path.join("naomi", "settings", "definitions", "README.md"), "r", encoding="utf-8") as fh:
        long_description += fh.read()

    setup(
        name='naomiutils',
        version='0.1',
        description='Code libraries for working with Naomi ROMs and EEPROMs.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='DragonMinded',
        author_email='dragonminded@dragonminded.com',
        license='Public Domain',
        url='https://github.com/DragonMinded/netboot',
        packages=[
            # Package for 3rd party.
            'naomi',
            'naomi.settings',
            # Include settings definitions.
            'naomi.settings.definitions',
            # Include default trojan.
            'homebrew.settingstrojan',
        ],
        package_data={
            # Make sure mypy sees us as typed.
            "naomi": ["py.typed", "README.md"],
            "naomi.settings": ["py.typed", "README.md"],
            # Make sure to include all existing settings.
            "naomi.settings.definitions": ["*.settings", "README.md"],
            # Make sure to actually include the trojan data.
            "homebrew.settingstrojan": ["settingstrojan.bin"],
        },
        python_requires=">=3.6",
    )
